import numpy as np

from epics import caget

import iexcode.instruments.cfg as iex
from iexcode.instruments.utilities import take_closest_value
from iexcode.instruments.xrays import getE, slit_get
from iexcode.instruments.VLS_PGM import mono_grating_get
from iexcode.instruments.electron_analyzer import EA, resolution_EA, getSESslit
from iexcode.instruments.ARPES import ARPES_extra_pvs
#############################################################################################################
##############################             Resolution              ##############################
#############################################################################################################
def resolution():
    """
    Calculate the theoretical resolution for the current beamline settings:
        ARPES: total resolution i.e. sqrt(kbT^2 + analyzer^2 + BL^2); default SES slit = 5.
        Kappa:  beamline contribution only
    """
    branch = iex.BL.branch
    grt = mono_grating_get()
    hv_eV = getE()
    slit_size = take_closest_value([10,20,50,100,200],round(slit_get(),0))

    if branch == "c":
        slit_SES = getSESslit()
        PE = int(EA.PassEnergy)
        Tsample = caget(ARPES_extra_pvs['TA'])
        resolution_ARPES(grt,hv_eV,slit_size,PE,slit_SES,Tsample,verbose=True)
        print("Calculate: resolution_ARPES(grt,hv_eV,slit_size,PE,slit_SES,Tsample)")
    else:
        resolution_beamline(grt,hv_eV,slit_size,verbose=True)
        print("Calculate: resolution_beamline(grt,hv_eV,slit_size)")



def resolution_calculate_beamline(grt,hv_eV,slit_size):
    """
    
    Previously: Calc_BL_Resolution
    """
    res_function={}
    res_function[10]  = {"MEG":(-0.000800494,2.56761e-05,1.83724e-08,-1.69223e-12,0),          "HEG":(-0.000414482,1.07456e-05,8.79034e-09,-8.42361e-13,0)}
    res_function[20]  = {"MEG":(-0.00156643, 3.16894e-05,2.72121e-08,-2.63642e-12,0),          "HEG":(-0.00054591, 1.35647e-05,2.01775e-08,-4.30789e-12,5.70112e-16)}
    res_function[50]  = {"MEG":(-0.00251543, 4.89022e-05,7.70055e-08,-1.66358e-11,2.21272e-15),"HEG":(-0.00173081, 2.97625e-05,4.90646e-08,-1.0706e-11, 1.43011e-15)}
    res_function[100] = {"MEG":(-0.00545563, 9.14928e-05,1.51335e-07,-3.30332e-11,4.41314e-15),"HEG":(-0.00360316, 5.83265e-05,9.76881e-08,-2.13767e-11,2.85877e-15)}
    res_function[200] = {"MEG":(-0.0111658,  0.000179785,3.01277e-07,-6.59309e-11,8.81755e-15),"HEG":(-0.00728107, 0.000116055,1.95149e-07,-4.27331e-11,5.71638e-15)}
    try:
        a,b,c,d,e = res_function[slit_size][grt]
        resolution = a + b*hv_eV + c*hv_eV**2 + d*hv_eV**3 + e*hv_eV**4

    except KeyError:
        print("WARNING: Slit size not listed, please choose one of the following:")
        print("        10, 20, 50, 100, 200 um.")
        resolution=0

    return resolution*1000


def resolution_mono_noise(grt,hv_eV):
    """
    jitter/noise in the mono contribution to the resolution 

    Previoulsy: Resolution_Mono
    """
    if grt == "MEG":
        K0=-2.3836
        K1=0.02083
        K2=7.8886e-06
    if grt == "HEG":
        K0=-2.0984
        K1=0.011938
        K2=3.6397e-06
    noise = K0 + K1*hv_eV + K2*hv_eV**2
    return noise


def resolution_beamline(grt,hv_eV,slit_size,verbose=True):
    """
    resolution of the beamline optics
    
    Resolution_BL
    """
    mono_noise = resolution_mono_noise(grt,hv_eV)
    theortical = resolution_calculate_beamline(grt,hv_eV,slit_size)
    resolution=np.sqrt(theortical**2 + mono_noise**2)
    if verbose:
        print("BL   : "+"%.0f" % theortical, "meV")
        print("Mono : "+"%.0f" % mono_noise, "meV")
        print("Total: "+"%.0f" % resolution, "meV")
    return resolution 


def resolution_ARPES(grt,hv_eV,slit_size,PE,slit_SES,T,verbose=True):
    """
    resolution of the ARPES
    """
    bl = resolution_beamline(grt,hv_eV,slit_size,verbose=False)
    KbT = T*25/300
    SES = resolution_EA(PE,slit_SES)
    
    resolution = np.sqrt(bl**2 + KbT**2 + SES**2)

    if verbose:
        print("BL + mono  : "+"%.0f" % bl, "meV")
        print("SES  : "+"%.0f" % SES, "meV")
        print("KbT  : "+"%.0f" % KbT, "meV")
        print("Total: "+"%.0f" % resolution, "meV")

    return resolution

def resolution_BL_eff(grating,hv,slit_3C,PE,slit_SES,T,Ef):
    M = resolution_mono_noise(grating,hv)
    KbT = T*25/300
    BL = resolution_calculate_beamline(grating,hv,slit_3C)
    SES = resolution_EA(PE,slit_SES)
    resolution = np.sqrt(BL**2+SES**2+KbT**2+M**2)
    effective = np.sqrt(Ef**2-SES**2-KbT**2-M**2)
    print("BL_th : "+"%.0f" % BL, "meV            SES   : "+"%.0f" % SES, "meV")
    print("Mono  : "+"%.0f" % M, "meV            KbT   : "+"%.0f" % KbT, "meV")
    print("Total : "+"%.0f" % resolution, "meV            Fermi : "+"%.0f" % Ef, "meV")
    print("BL_eff: "+"%.0f" % effective, "meV")


def resolution_BL_eff2(grating,hv,slit_3C,PE,slit_SES,T,Ef,Dirty_Au):
    M = resolution_mono_noise(grating,hv)
    KbT = T*25/300
    BL = resolution_calculate_beamline(grating,hv,slit_3C)
    SES = resolution_EA(PE,slit_SES)
    resolution = np.sqrt(BL**2+SES**2+KbT**2+M**2+Dirty_Au**2)
    effective = np.sqrt(Ef**2-SES**2-KbT**2-M**2-Dirty_Au**2)
    print("BL_th : "+"%.0f" % BL, "meV            SES   : "+"%.0f" % SES, "meV")
    print("Mono  : "+"%.0f" % M, "meV            KbT   : "+"%.0f" % KbT, "meV")
    print("Total : "+"%.0f" % resolution, "meV            Fermi : "+"%.0f" % Ef, "meV")
    print("BL_eff: "+"%.0f" % effective, "meV            Sample : "+"%.0f" % Dirty_Au, "meV")

