"""
functions related to using x-rays

"""
import numpy as np
#import math
from time import sleep

from epics import caget,caput

import iexcode.instruments.cfg as iex
from iexcode.instruments.IEX_cameras import _enable_endstation_cameras
from iexcode.instruments.current_amplifiers import ca_average, ca_detector_list
from iexcode.instruments.diagnostics import *
from iexcode.instruments.FMB_mirrors import FMB_mirror_get
from iexcode.instruments.IEX_VPU import *
from iexcode.instruments.Logfile import log_update
from iexcode.instruments.m3r import m3r_align,m3r_branch,m3r_switch_branch,m3r_position_dictionary
from iexcode.instruments.shutters import *
from iexcode.instruments.slits import *
from iexcode.instruments.utilities import print_warning_message,make_table
from iexcode.instruments.valves import *
from iexcode.instruments.VLS_PGM import *

mpa = iex.mpa

##############################################################################################################
##############################            resets and detector lists         ##############################
##############################################################################################################
def _xrays_reset(verbose=True):
    if verbose:
        print("resetting the mono limits")

    mono_limits_reset()

    if verbose:
        print("resetting beamline encoders")

    xray_motor_encoder_sync()


def _xrays_detector_list():
    """
    defualt detectors for the APRES endstation
    used by: 
         CA_Live_StrSeq()
        Detector_Triggers_StrSeq()
        BeforeScan_StrSeq() => puts everybody in passive
        CA_Average()
    WARNING: can't have more than 5 otherwise CA_Live_StrSeq gets angry.

    Previously: part of Detector_List
    """
    ca_list=[["b",4],["b",13]]

    return ca_list


def _xrays_detector_dictionary():
    """
    returns a dictionary of the default detectors
    
    Previously: Detector_Default
    """
    det_dict={}
    sr_main_shutter={
        1:"S:SRcurrentAI.VAL",
        2:"EPS:29:ID:SS1:POSITION"
        }
    bl_energy={
        3:"29idmono:ENERGY_MON",
        4:"ID29:EnergySet.VAL",
        5:"ID29:Energy.VAL"
        }
    bl_keithleys={
        6:"29idb:ca1:read",
        7:"29idb:ca2:read",
        8:"29idb:ca3:read",
        9:"29idb:ca4:read",
        10:"29idb:ca5:read",
        11:"29idb:ca9:read",
        12:"29idb:ca12:read",
        13:"29idb:ca13:read"
        }
    det_dict.update(sr_main_shutter)
    det_dict.update(bl_energy)
    det_dict.update(bl_keithleys)
    return det_dict

##############################################################################################################
##############################            get beamline info         ##############################
##############################################################################################################

def xrays_get_all(verbose=True):
    """
    gets info about current beamline setting
    returns a dictionary

    """
    vals={}

    if verbose:
        print("-----------------------------------------------------------")   
    vals.update(ID_get_all(verbose=verbose))
    vals.update(mono_get_all(verbose=verbose))
    
    if verbose:
        print("-----------------------------------------------------------")
    vals.update(slits_get_all(verbose=verbose))
    vals.update({"M0":FMB_mirror_get(0,verbose=verbose)})
    vals.update({"M1":FMB_mirror_get(1,verbose=verbose)})
    vals.update({"M3":FMB_mirror_get(3,verbose=verbose)})
    if verbose:
        print("-----------------------------------------------------------")

    return vals

def _xrays_log_entries(**kwargs):
    """
    Writes CSV file for the MDA scans with the following default parameters:
    FilePath='/home/beams/29IDUSER/Documents/User_Folders/UserName/'
    Filename is set by logname_set(FileName=None), default:YYYYMMDD_log.txt, where YYYYMMDD is when folders were set
        (FileName=None,FileSuffix=None) is they are specified then it will use those values instead

    comment="Additional comments, last column"

    Previously: scanlog
    """
    #default parameters
    kwargs.setdefault('comment','')
    #global BL

    ioc = iex.BL.ioc

    #scan Num
    scan = iex.BL.mda.prefix(ioc)+str(iex.BL.mda.fileNum(ioc))
    entry_list=["scan"]
    pv_list=[scan]
    format_list=["s" ]
    #scan info
    drive = caget(ioc+"scan1.P1PV")
    start = caget(ioc+"scan1.P1SP")
    stop  = caget(ioc+"scan1.P1EP")
    step  = caget(ioc+"scan1.P1SI")
    entry_list.append('drive','start','stop' ,'step')
    pv_list.append(drive,start,stop ,step)
    format_list.append("s",".3f",".3f",".3f")

    #endstation info
    if iex.BL.endstation_name == 'ARPES':
        endstation_entry, endstation_pv, endstation_format = iex.BL.log_entries(**kwargs)
        entry_list.append(endstation_entry)
        pv_list.append(endstation_pv)
        format_list.append(endstation_format)
    elif iex.BL.endstation_name == 'Kappa':
        endstation_entry, endstation_pv, endstation_format = iex.BL.log_entries(**kwargs)
        entry_list.append(endstation_entry[:-7])
        pv_list.append(endstation_pv[:-7])
        format_list.append(endstation_format[:-7])

    #beamline info
    ID_mode = ID_mode_get(verbose=False) if iex.BL.xrays else "no_xrays"
    ID_QP_ratio = ID_QP_ratio_get(verbose=False)[0] if iex.BL.xrays else "no_xrays"
    ID_sp  = round(ID_SP_get_eV()) if iex.BL.xrays else 0
    ID_rbv = round(ID_rbv_get_eV,4) if iex.BL.xrays else 0
    hv = mono_energy_get() if iex.BL.xrays else 0
    grt = mono_grating_get() if iex.BL.xrays else 0
    slit_size = slit_get()[0] if iex.BL.xrays else 0
    entry_list.append("hv","exit_slit","GRT", "ID_SP", "ID_RBV", "ID_Mode", "ID_QP")
    pv_list.append(hv,slit_size,grt, ID_sp, ID_rbv,ID_mode, ID_QP_ratio)
    format_list.append(".2f",".0f","s",".1f",".1f","s", ".0f")

    #endstation info 2: Fanny can I change to order to get rid of this complication?
    if iex.BL.endstation_name == 'Kappa':
        endstation_entry, endstation_pv, endstation_format = iex.BL.log_entries(**kwargs)
        entry_list.append(endstation_entry[-7:])
        pv_list.append(endstation_pv[-7:])
        format_list.append(endstation_format[-7:])

    #timestamp and comments
    t = time.strftime("%D-%H:%M:%S")
    comment = kwargs["comment"]
    entry_list.append('time','comment')
    pv_list.append(t,comment)
    format_list.append("s","s")

    try:
        log_update(iex.BL.endstation,iex.BL.ioc,entry_list,pv_list,format_list)
    except:
        print("scanlog did not write to file, check for errors.")
        
##############################################################################################################
##############################            beamline encoders         ##############################
##############################################################################################################

def xray_motor_encoder_sync():
    ioc = "29idb:"
    for motor in [13,14,15,16,26,27]:
        pv = ioc+"m"+str(motor)+".SYNC"
        caput(pv,1)
        print("synch encoder: "+pv)

##############################################################################################################
##############################               energy : ID + mono           ##############################
##############################################################################################################
def energy_get_all(verbose=True):
    """
    returns ID_mode,ID_QP_ratio,ID_sp,ID_rbv,hv,grt 

    """
    d = ID_get_all(verbose=False)
    d.update(mono_get_all(verbose=False))
    keylist = ['ID_mode', 'ID_QP_ratio', 'ID_sp', 'ID_rbv','hv','grating']
    values = []
    for key in keylist:
        values.append(d[key])
        if verbose:
            print(key,d[key])

    return tuple(values)
    
def energy_get():
    """
    return the mono energy readback
    """
    hv = mono_energy_get()
    return hv

def energy(hv_eV,slit_coeff=1,m3r=True,QP_ratio=100,verbose=True):
    """
    sets the ID for optimal flux based on calibration
    sets the mono
    sets the beamline aperture/slits

    slit_coeff < 1 is used to kill flux by closing the aperature after the mono (slit-2B)
    
    m3r => if True optimizes the mirror for the d-branch only
    Previously: Set_BL, energy
    """
    if hv_eV != _energy_range_check(hv_eV):
        message_string = 'request photon energy '+str(hv_eV)+' not with the allowed range'
        message_string = '\n closest allowed energy is '+str(_energy_range_check(hv_eV))
        print_warning_message(message_string)


    ID_energy_set_eV(hv_eV,QP_ratio,verbose=verbose)
    mono_energy_set(hv_eV,verbose=verbose)
    apertures_set(c_2B=slit_coeff,c_1A=1,verbose=verbose)

    if m3r == True:
        m3r_align()

def _energy_range_min_max():
    """
    returns the min,max energies for the current grating/ID setting
    """
    ID_min, ID_max = ID_energy_range_eV()
    mono_min, mono_max = mono_energy_range()
    hv_max = min(ID_max, mono_max)
    hv_min = max(ID_min, mono_min)
    return hv_min,hv_max

def _energy_range_check(hv):
    """
    checks if the energy is within the allowed range for both the mono
    and the ID. If outside the ranges set it to the closest allowed value

    Previously: SetRange()
    """
    hv_min,hv_max = _energy_range_min_max()
    hv = min(hv_max,hv)
    hv = max(hv_min,hv)
    return hv

def qp(QP_ratio):
    """
    Used to set the QP (quasiperiodicity) of the ID for harmonic rejection
    QP_ratio =>
         100 fully periodic; max flux
         70 minimum allowed value, most quasiperiodic

    After changing QP, you will need to set
        polarization
        energy
    """
    ID_QP_mode_set(QP_ratio)
##############################################################################################################
################################             beamline slits  = apertures          ##############################
############################################################################################################## 
def aperture_fit(hv,slit_num):
    """
    used close the beamline apertures/slits to only take the center of the beam, 
        i.e. no heat bump (determined emperically by looking at the the shift in energy vs slit position)

    Previously Aperture_Fit
    """
    K=slit_coef(slit_num)[1]
    sizeH=K[0]+K[1]*hv+K[2]*hv*hv
    sizeV=K[3]+K[4]*hv+K[5]*hv*hv
    return [round(sizeH,3),round(sizeV,3)]

def apertures_set(c_2B=1,c_1A=1,verbose=True):
    """
    Sets slit-1A and slit-2B for the current 
    grating and photon energy to remove the heat bump

    c_1A and c_2B are used to take a different ratio of the beam 
    (1 = normal ops, < 1 to reduce flux)

    Previously: SetSlit_BL
    """
    hv_rbv = mono_energy_get()
    grt  = mono_grating_get()
    
    # slit were defined for the range: 500 - 2000
    hv = max(hv_rbv,500)
    hv = min(hv_rbv,2000)
    c = 4.2/2.2 # to account for the magnification difference between gratings

    if grt == 'MEG':
        V=0.65        #  set to 65% of RR calculation for both grt => cf 2016_2_summary
    elif grt=='HEG':
        V=0.65*c        #  set to 65% of RR calculation (no longer 80%) => cf 2016_2_summary

    try:
        slit_position = read_dict('Dict_Slit.txt')
    except KeyError:
        print_warning_message("Unable to read dictionary")
        return

    V2_center= slit_position[grt]['S2V']
    H2_center= slit_position[grt]['S2H']
    V1_center= slit_position[grt]['S1V']
    H1_center= slit_position[grt]['S1H']

    Size1A=( aperture_fit(hv,1)[0]*c_1A,       aperture_fit(hv,1)[1]*c_1A )
    Size2B=( aperture_fit(hv,2)[0]*c_2B, round(aperture_fit(hv,2)[1]*c_2B*V,3))
    
    slit1A_set(Size1A[0],Size1A[1],center=(H1_center,V1_center),verbose=verbose)    # standard operating
    slit2B_set(Size2B[0],Size2B[1],center=(H2_center,V2_center),verbose=verbose)


def mvID(val,QP_ratio=100):
    """
    Sets the ID absolute set point (not optimized , mono & apertures stay fixed).
    
    to go to most flux use energy
    """
    ID_SP_set(val,QP_ratio=100)

def mvID_calibrated(val,QP_ratio=100):
    """
    Sets the ID peak to val (calibrated).

    """
    ID_energy_set(val,QP_ratio)


def scan_ID(ID_sp_start,ID_sp_stop,ID_sp_step,**kwargs):
    """
    scan the ID set point
    """
    val_pv,rbv_pv = ID_scan_pvs()
    iex.BL.mda.fillin(val_pv,rbv_pv,ID_sp_start,ID_sp_stop,ID_sp_step,**kwargs)


def mvmono(val):
    """
    Sets the mono energy (insertion device & apertures stay fixed).
    """
    mono_energy_set(val)

def getE():
    """ 
    Returns current mono set point.
    """
    return energy_get()

def grating(grating_name):
    """
    used to switch between gratings
        grating = 'HEG' / 'MEG'

        check current grating 
        closes the main shutter 
        sets the mono energy 

    Previously: Switch_Grating
    """
    current_grt = mono_grating_get()
    if grating != current_grt:
        shutter_open = main_shutter_status()
        main_shutter_close()
        print("Switching grating, please wait...")
        mono_grating_change(grating_name,verbose=False)
        apertures_set()

        if shutter_open:
            main_shutter_open()
    
        print(dateandtime())
    else: 
        print("grating: "+grating)

def grating_get():
    """
    returns the current grating
    """
    grating_name = mono_grating_get()
    return grating_name


def polarization(ID_mode):
    """
    Change beam polarization: ID_mode = 'H', 'V', 'RCP' or 'LCP'
    """
    ID_switch_mode(ID_mode)

def polarization_get():
    """
    gets the current polarization
    """
    ID_mode_get()

def scanmono(start,stop,step,average_pnts=1,**kwargs):
    """
    scans the mono at the current ID value

    """
    ca_average(average_pnts,ca_detector_list(iex.BL.branch))

    mono_scan_fillin(iex.BL.mda,start,stop,step,**kwargs)

    mono_energy_set(start)
    iex.BL.mda.go(**kwargs)

    mono_scan_after(iex.BL.mda)



def scanXAS(start_stop_step_lists,ID_eV=None,ID_tracking=False, **kwargs):
    """
    scans the beamline energy:
        ID_eV = None        => the ID stays at the current value (default) 
        ID_eV = value       => set the ID 
        ID_track = True     => ID will follow with the mono (SLOW but required in ARPES)
      
    start_stop_step_lists is a list of lists for the different scan ranges
        start_stop_step_lists = [[start1,stop1,step1], [start1,stop1,step1],...]
        Note duplicates are removed and the resulting array is sorted in ascending order

    **kwargs:
        scan_dim = 1 (default)
        average_pnts: if using Keithlys this set the averaging; None for no averaging
        m3r: if True the mirror is optimized before the start of the scan
        mcp
        execute: True/False to start the scan => True (default)
        QP_ratio: if other than fully periodic
    
    Normalization:
        - If in the D-branch the Mesh is put in but is not removed
        - If in the C-branch we use the slit blades for normalization (ca13)
    """
    kwargs.setdefault("scan_dim",1)
    kwargs.setdefault("average_pnts",1)
    kwargs.setdefault("m3r",True)
    kwargs.setdefault("mcp",True)    
    kwargs.setdefault('execute',True)   

    scan_dim=kwargs['scan_dim']
    
    #Setting up the ScanRecord for Mono and ID in Table mode
    mono_array,ID_array = BL_energy_tables(start_stop_step_lists,**kwargs)
    kwargs.update('positioner_num',1)
    mono_scan_fillin_table(iex.BL.mda,scan_dim,mono_array,**kwargs)
    if ID_tracking == True:
        kwargs.update('positioner_num',2)
        ID_scan_fillin_table(iex.BL.mda,scan_dim,ID_array,**kwargs)
        energy(ID_array[0],m3r=kwargs["m3r"])
    elif ID_eV != None:
        #Setting the beamline energy
        energy(ID_eV,m3r=kwargs["m3r"])

    #Averaging and Normalization
    ca_average(kwargs['average_pnts'],ca_detector_list(iex.BL.branch))
    if iex.BL.branch=="d":
        mesh_d("In")
        print_warning_message('MeshD is not automatically removed')

    #mpa
    if iex.BL.endstation == "Kappa" and kwargs["mcp"]:
        mpa.HV_on()

    #Scanning
    if kwargs['execute']:
        mono_energy_set(mono_array[0])
        iex.BL.mda.go(scan_dim)
    
        #Setting everything back
        mono_energy_set(ID_eV)
        mono_scan_after(iex.BL.mda,scan_dim)
        iex.BL.mda.table_reset_after(scan_dim)

        if iex.BL.endstation == "Kappa": 
            if kwargs["mcp"]: 
                mpa.HV_off()
            print("WARNING: Mesh"+iex.BL.branch+" is still In")        


def BL_energy_tables(start_stop_step_lists,**kwargs):
    """
    returns mono_array and ID_array for BL energy scans
    *energies:
        start,stop,step
        ListofLists, e.g.{[400,420,1],[420,425,0.25]...}
    **kwargs:
        QP_ratio = 100; completely periodic, normal operations
    Previously: Tables_BLenergy
    """
    kwargs.setdefault('QP_ratio', 100)
    mono_array = make_table(start_stop_step_lists)
    ID_array=np.array([])
    ID_mode,ID_QP_ratio,ID_sp,ID_rbv,hv,grt = energy_get_all(verbose=False)
    for hv_eV in mono_array:
        ID_array=np.append(ID_array,ID_calc_SP(grt,ID_mode,hv_eV))
    return mono_array,ID_array


##############################################################################################################
###########################               switch branch                 ######################
##############################################################################################################


def switch_branch(branch, force=False, shutter=True,scan_reset=True,enable_cams=True):
    """Switch beam into which = \"c\" or \"d\" branch (by retracting/inserting deflecting mirror)
        Optionnal keyword argument:
            force: change m3r even if the beamline think it is already there (default=False)
            shutter: open the branch shutter, does not open the valve (default=False) 
            reset: resets the scanRecord and the global BL for the selected branch
            enable_cams: to turn on/off branch camera  

    Previously: Switch_Branch
    """
    branch = branch.lower()
    # Check current branch:
    d = m3r_position_dictionary()
    if branch in d.keys():
        if m3r_branch() == branch and not force:
            m3r_position = FMB_mirror_get(3,verbose=False)
            if np.sum(np.array(m3r_position)):
                print("\nMirror homed...")
                print("...Try again using: Switch_Branch(which, forced=True)")
            else:
                print("\nWell, looks like the beam is already in this branch...")
                print("...if you want to force the mirror motion, use the argument: forced=True")
        else:
            # Close both shutters:
            print("\n")
            print("Switching branch, please wait...")
            if shutter:
                branch_shutter_close('c')
                branch_shutter_close('d')
    
            #Move M3R:
            m3r_switch_branch(branch)
            print(dateandtime()," -  In "+branch+"-Branch")
            sleep(2)

        # Open branch shutters:
            if shutter:
                branch_shutter_open(branch,valve=False)

            if scan_reset:
                iex.BL.mda.reset()

            if enable_cams:
                _enable_endstation_cameras(branch)

    else:
        print_warning_message(branch+' is not a valid branch selection')
#             if not nocam:

##############################################################################################################
###########################                      mirrors                    ######################
##############################################################################################################
def get_mirrors():
    for mirror_num in [0,1,3]:
        FMB_mirror_get(mirror_num,verbose=True)


##############################################################################################################
###########################                  slits and resolution                    ######################
##############################################################################################################
def slit(size,verbose=True):
    """
    sets the exit slit based on iex.BL.endstation

        ARPES = 0 < x < 300  um
        Kappa  = 0 < x < 1000 um
    """
    branch = iex.BL.branch
    exit_slit(branch,size,verbose=verbose)


def slit_get(verbose=True):
    """
    sets the exit slit based on iex.BL.endstation

        ARPES = 0 < x < 300  um
        Kappa  = 0 < x < 1000 um
    """
    exit_slit_get(branch,verbose=verbose)
    return slit_size,slit_center





