from numpy import nan

from epics import caput, caget

import iexcode.instruments.cfg as iex
from iexcode.instruments.current_amplifiers import Keithley,SRS
from iexcode.instruments.utilities import print_warning_message
##############################################################################################################
################################            default positions             ##############################
##############################################################################################################
def _diagnostics_presets():
    presets={
        'H-wire':['OUT','Start'],
        'V-wire':['OUT','Start'],
        'H-Diagon':['OUT','Si-2800','ML-400','ML-460'],
        'V-Diagon':['OUT','Si-2800','ML-400','ML-460'],
        'D1A':['OUT','Mesh'],
        'D2B':['OUT','Mesh','YAG'],
        'D3B':['OUT','YAG'],
        'D4C':['OUT','Mesh','YAG'],
        'gas-cell':['OUT','Diode1','Diode2'],
        'D4D':['OUT','Mesh','YAG'],
        'D5D':['OUT','Mesh','YAG','Diode'],
    }
    return presets

#For new endstation modify here:
def _diagnostics_inout_dict():
    """
    dictionary to convert quick_name to preset_names
    """
    #quick_name: (diag_name, preset_in,preset_out)
    d={
        'mesh_W':('D1A','Mesh','OUT'),
        'mesh_c':('D4C','Mesh','OUT'),
        'mesh_d':('D4D','Mesh','OUT'),
        'diode_c':('gas-cell','Diode1','OUT'),
        'diode_d':('gas-D4D','Diode','OUT'),
        }

    return d

#For new endstation modify here:
def _diagnostic_CA_dict():
    CA_dict={
        'diode_c':Keithley('b',15),
        'mesh_d':SRS("29idMZ0:scaler1.S14", '29idd:A4'),
    }
    return CA_dict

##############################################################################################################
################################            access presets            ##############################
##############################################################################################################
def _diagnostics_preset_pv(diag_name,preset_name):
    """
    dictionary of the motor pvs for the beamline diagnostics
    """
    ioc = '29idb:'
    d = _diagnostics_presets()
    pv=None
    
    if diag_name.casefold() in (key.casefold() for key in list(d.keys())):
        idx = [item.lower() for item in list(d.keys())].index(diag_name.lower())
        diag_name=list(d.keys())[idx]
        if preset_name.casefold() in (preset.casefold() for preset in d[diag_name]):
            idx = [item.lower() for item in d[diag_name]].index(preset_name.lower())
            preset_name = d[diag_name][idx]
            pv = ioc+diag_name+'_'+preset_name
        else:
            message ='preset_name = '+preset_name+' not a valid name'
            message +='\npreset_names:'+str(d[diag_name])
            print_warning_message(message)
    
    else:
        message = 'diag_name = '+diag_name+' not a valid name'
        message +='\npreset_names:'+str(list(d.keys()))
        print_warning_message(message)
    
    return pv

def diagnostics_go2(diag_name,preset_name):
    pv = _diagnostics_preset_pv(diag_name,preset_name)
    caput(pv+'_Trigger.PROC',1)

def diagnostics_sp_read(diag_name,preset_name):
    pv = _diagnostics_preset_pv(diag_name,preset_name)
    caget(pv+'_Setpoint.VAL')

def _diagnostics_sp_write(diag_name,preset_name):
    pv = _diagnostics_preset_pv(diag_name,preset_name)
    caget(pv+'_Setpoint.VAL')
    

##############################################################################################################
################################            quick names            ##############################
##############################################################################################################
def diagnostics_list_all(verbose=True):
    """
    lists all diagnostics
    """
    d = _diagnostics_presets()
    if verbose:
        for key in d.keys():
            print(key)
    return list(d.keys())

def diagnostic(name,in_out):
    '''
    Inserts/retracts a diagnostic(motor number or name) either = "In" or "Out"
    
    '''
    d_quick = _diagnostics_inout_dict()
    if name in d_quick.keys():
        diag_name, preset_in,preset_out = d_quick[name]
        if in_out.lower() == "in":
            preset_name = preset_in
        elif in_out.lower() == "out":
            preset_name = preset_in
        diagnostics_go2(diag_name,preset_name)
        print("\n"+name+": "+ in_out)
    else:
        message = 'name = '+name+' is not a valid diagnostic name'
        message +='\n valid names are '+diagnostics_list_all(verbose=False)
        print_warning_message(message)

def diagnostics_all_out(diode_stay_in=False,mesh_stay_in=False):
    """
    Retracts all diagnostic unless
        diode_stay_in = True => leaves diode_c/diode_d In
        mesh_stay_in = True => leaves mesh_c/mesh_d In

    Previously: AllDiagOut
    """
    message = "All diagnostics out"
    d_quick = _diagnostics_inout_dict()

    diag_list = list(_diagnostics_presets().keys())

    #if staying remove diag_name from list
    if diode_stay_in:
        branch = iex.BL.branch
        if branch == 'c':
            diode_stay_name = d_quick['diode_c'][0]
            message += '\ndiode_c => In'
        elif branch == 'd':
            diode_stay_name = d_quick['diode_d'][0]
            message += '\ndiode_d => In'
        diag_list.pop(diode_stay_name) 

    if mesh_stay_in:
        branch = iex.BL.branch
        if branch == 'c':
            mesh_stay_name = d_quick['mesh_c'][0]
            message += '\nmesh_c => In'
        elif branch == 'd':
            mesh_stay_name = d_quick['mesh_d'][0]
            message += '\nmesh_d => In'
        diag_list.pop(mesh_stay_name) 

    #moving the diagnostics in list
    for diag_name in diag_list:
        diagnostics_go2(diag_name,'OUT')
    
    #putting stays in if not already in
    if diode_stay_in:
        diagnostics_go2(diode_stay_name,'Diode')
    if mesh_stay_name:
        diagnostics_go2(mesh_stay_name,'Mesh')

    print("\n",message)

def mesh_W(In_Out):
    """
    Inserts/retracts RSXS mesh (post-slit); arg = \"In\" or \"Out\"

    Previously MeshW
    """
    diag=_diagnostics_dict()
    motor=5; position=diag[In_Out][motor]
    caput("29idb:m"+str(motor)+".VAL",position,wait=True,timeout=18000)
    print("\nD1A mesh_W: "+ In_Out)


def diode_c(In_Out):
    """
    Inserts/retracts ARPES (gas-cell) diode; arg = \"In\" or \"Out\"
    
    Previously: DiodeC
    """
    diag=_diagnostics_dict()
    motor=20; position=diag[In_Out][motor]
    caput("29idb:m"+str(motor)+".VAL",position,wait=True,timeout=18000)
    print("\nARPES Diode: "+ In_Out)


def diode_d(In_Out):
    """
    Inserts/retracts RSXS diode; arg = \"In\" or \"Out\"

    Previously:DiodeD
    """
    diag=_diagnostics_dict()
    motor=28; position=position=diag[In_Out][motor]
    if type(position) == list:
        position=position[1]
    caput("29idb:m"+str(motor)+".VAL",position,wait=True,timeout=18000)
    print("\nRSXS Diode: "+ In_Out)


def mesh_d(In_Out):
    """
    Inserts/retracts RSXS mesh (post-slit); arg = \"In\" or \"Out\"
    
    Previoulsy: MeshD
    """
    diag=_diagnostics_dict()
    motor=28; position=position=diag[In_Out][motor]
    if type(position) == list:
        position=position[0]
    caput("29idb:m"+str(motor)+".VAL",position,wait=True,timeout=18000)
    print("\nD5D Au-Mesh: "+ In_Out)

def diagnostic_read(diode_name, verbose=True):
    """
    reads the current amplifier and returns the value
    """ 
    try:
        det = _diagnostic_CA_dict[diode_name]
        val=det.get(verbose)
    except:
        diode_pv="not connected"
        val = nan
        print('pv not defined')

    if verbose:
        print(diode_name,val, "(pv = "+diode_pv+")")
    return val

def diode_c_read(verbose=True):
    """
    reads the current amplifier and returns the value
    quiet = False to pring
    """ 
    val = diagnostic_read('diode_d',verbose)
    return val

def diode_d_read(verbose=True):
    """
    reads the current amplifier and returns the value
    quiet = False to pring
    """ 
    val = diagnostic_read('diode_d',verbose)
    return val

def mesh_d_read(verbose=True):
    """
    reads the current amplifier and returns the value
    quiet = False to pring
    """ 
    val = diagnostic_read('diode_d',verbose)
    return val


###############################################################################################
####################################### FLUX CONVERSION #######################################
###############################################################################################




